"""File with version and history information."""

__version__ = '1.1.0'

"""
History Information:
1.0.0
    First Version
1.0.1
    Removed tests folder from pip
1.0.2
    Fixed documentation
1.0.3
    Fixed documentation
1.0.4
    Fixed issue #4
1.1.0
    Corrected version number.
    Previous should have been 1.1.0 as it changes function behaviour.
1.2.0
    Moved debug functionality to a separated file.
    Debug output now includes curl command.
"""
